package function

import (
	"encoding/json"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"

	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
	"gitlab.smartblocklab.com/smartcontracts/transactionCC/mngr"
)

// Функция блокировки средств на счете пользователя
// userParameter - параметр пользователя полученный с UI
// systemParameter - системный параметр полученный с backend-cryptoenter
// manager - интерфейс для работы с блокчейном
func LockFundsOnAccount(userParameter *requests.Parameter, systemParameter *requests.Parameter, manager chaincode.IManager) (interface{}, error) {
	var txn transactionCC.TxnFundsLockInput
	// Проверяем что преобразовывание байтов из параметров переданных с ui к объекту TxnFundsLockInput происходит без ошибок
	err := json.Unmarshal(userParameter.Param, &txn)
	if err != nil {
		return nil, errors.Errorf("Can't unmarshal userParam to TxnFundsLockInput. Error: %s", err)
	}

	// Проверка входящего пользователя на то, что данный параметр не пустой
	if txn.InUser == nil {
		return nil, errors.Errorf("In user can't by null or empty")
	}

	// Проверка исходящего пользователя на то, что данный параметр не пустой
	if txn.OutUser == nil {
		return nil, errors.Errorf("Out user can't by null")
	}

	// Проверка исходящего пользователя на то, что длина строки userId данного пользователя не равняется нулю
	if len(txn.OutUser.UserID) == 0 {
		return nil, errors.Errorf("Out user can't by empty")
	}

	// Проверка, что запрос подписан userId соответствующий исходящему пользователю (который блокирует средства)
	if txn.OutUser.UserID != userParameter.UserID {
		return nil, errors.Errorf("Can't %s because signer UserID %s not equals with locked founds UserID %s", invoke.LockFundsOnAccountFunction, userParameter.UserID, txn.OutUser.UserID)
	}

	// Проверка, что входящий и исходящий userId не равны при операции обмена с комиссией с обоих пользователей
	if txn.OutUser.UserID == txn.InUser.UserID &&
		txn.ProcessType == transactionCC.ExchangeWithBothCommission_ProcessType {
		return nil, errors.Errorf("%s. Transaction with processType 'ExchangeWithBothCommission' out user id {%s} and in user id {%s} can't be equals %s", invoke.LockFundsOnAccountFunction, txn.OutUser.UserID, txn.InUser.UserID)
	}

	// Если все проверки пройдены успешно, то определяем переменную lockFundsOnAccountArg
	var lockFundsOnAccountArg transactionCC.LockFundsOnAccountArg
	// Проверяем что преобразовывание байтов из параметров переданных с backend к объекту LockFundsOnAccountArg происходит без ошибок
	err = json.Unmarshal(systemParameter.Param, &lockFundsOnAccountArg)
	if err != nil {
		return nil, errors.Errorf("Can't unmarshal systemParams to LockFundsOnAccountArg. Error %s", err)
	}

	// Записываем параметры id операции и тип комиссии
	txn.OperationID = lockFundsOnAccountArg.OperationID

	// Отправляем все параметры в TransactionManager для блокировки средств
	return mngr.GetTransactionManager(manager).LockFundsOnAccount(&txn)
}
