package mngr

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
	"fmt"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/cryptocurrencyCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/system_user"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/model/commissionCC"
	"github.com/shopspring/decimal"
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
)

// isEnoughBalance check whenever there is enough balance
func (txnMngr *TransactionManagerImpl) isEnoughBalance(balance *nums.CurrencyNumber, amount *nums.CurrencyNumber) error {
	if amount == nil || amount.Num == decimal.NewFromFloat(0) {
		return errors.Errorf("amount for transfer can't be 0")
	}

	if balance == nil || balance.Num == decimal.NewFromFloat(0) {
		return errors.Errorf("balance can't be 0")
	}

	if balance.CurrencyType() != amount.CurrencyType() {
		return errors.Errorf("balance CurrencyType %s and amount CurrencyType %s not equals", balance.CurrencyType(), amount.CurrencyType())
	}

	isBalanceLessAmount, err := balance.Less(amount)
	if err != nil {
		return err
	}
	if isBalanceLessAmount {
		// We do not have enough money to transfer => returning the error
		return errors.Errorf("not enough funds to transfer money. required %s, while available %s", amount.RoundBankString(), balance.RoundBankString())
	}
	return nil
}

// запись информации о транзакции
func (txnMngr *TransactionManagerImpl) Add(mapCurrencyTypeKeys transactionCC.CurrencyTypeKeys, currencyType string, transactionID string) transactionCC.CurrencyTypeKeys {
	if len(mapCurrencyTypeKeys) == 0 {
		mapCurrencyTypeKeys = make(transactionCC.CurrencyTypeKeys)
	}

	keys, ok := mapCurrencyTypeKeys[currencyType]
	if !ok {
		keys = make(map[string]bool)
		mapCurrencyTypeKeys[currencyType] = keys
	}
	keys[transactionID] = true

	return mapCurrencyTypeKeys
}

// depositBitcoinFundsTransaction сохранение/создание транзакции на ввод криптовалюты
func (this *TransactionManagerImpl) depositBitcoinFundsTransaction(cryptocurrencyStore *cryptocurrencyCC.CryptocurrencyStore) ([]*transactionCC.TransactionRecord, error) {
	logging.Get().Infof("deposit funds transaction for cryptocurrencyStore %s", cryptocurrencyStore.ExternalTransactionID)

	if cryptocurrencyStore == nil {
		return nil, errors.Errorf("cryptocurrencyStore can't by null")
	}

	err := cryptocurrencyStore.Validate()
	if err != nil {
		return nil, logging.Get().Error(err.Error())
	}

	// проверяем задал ли пользователь конкретный внутренний счет на который нужно переводить стредства
	// по пользователю находим настройки
	userPreference, err := this.UserPreferenceCC.GetPreference(&requests.Id{Id: cryptocurrencyStore.InAccountInfo.UserID})
	if err != nil {
		return nil, err
	}

	internalAccountId := ""

	if userPreference != nil {
		externalAccountId := cryptocurrencyStore.InAccountInfo.AccountID
		// по внешнему номеру счета находим внутренний счет
		internalAccountId = userPreference.Fields[externalAccountId]
	}

	var internalUserAccountInfoForAddCryptocurrency *accountCC.Account
	// если пользователь не привязал внешний счет к внутреннему, тогда подтягиваем ему автоматический Главный Кошелек, Главный Счет
	if len(internalAccountId) == 0 {
		internalUserAccountInfoForAddCryptocurrency, err = this.AccountCC.CheckAndLoadAccount(&accountCC.UserAccountInfo{UserID: cryptocurrencyStore.InAccountInfo.UserID})
		if err != nil {
			return nil, err
		}
	} else {
		internalUserAccountInfoForAddCryptocurrency, err = this.AccountCC.CheckAndLoadAccount(&accountCC.UserAccountInfo{
			UserID: cryptocurrencyStore.InAccountInfo.UserID,
			TxnAccountInfo: &accountCC.TxnAccountInfo{
				AccountID: internalAccountId,
			},
		})
		if err != nil {
			return nil, err
		}
	}

	if len(internalUserAccountInfoForAddCryptocurrency.AccountID) == 0 {
		chaincode.GlobalLogger.Errorf("beginUserAccountInfo.AccountID for UserID %s is null", cryptocurrencyStore.InAccountInfo.UserID)
		return nil, err
	}

	chaincode.GlobalLogger.Infof("internalUserAccountInfoForAddCryptocurrency.AccountID: %s", internalUserAccountInfoForAddCryptocurrency.AccountID)
	chaincode.GlobalLogger.Infof("internalUserAccountInfoForAddCryptocurrency.WalletID: %s", internalUserAccountInfoForAddCryptocurrency.WalletID)
	chaincode.GlobalLogger.Infof("internalUserAccountInfoForAddCryptocurrency.UserID: %s", internalUserAccountInfoForAddCryptocurrency.UserID)

	operationId := fmt.Sprintf("%s.%s", cryptocurrencyStore.InAccountInfo.AccountID, cryptocurrencyStore.ExternalTransactionID)

	transactionAmountRecords, err := this.depositAmount(internalUserAccountInfoForAddCryptocurrency, cryptocurrencyStore.Amount, operationId)
	if err != nil {
		return nil, err
	}

	fromUser, err := this.UserCC.GetUser(&requests.Id{Id: internalUserAccountInfoForAddCryptocurrency.UserID})
	if err != nil {
		return nil, err
	}

	// создаем аккаунт для блокировки средств на пользователя с комиссией
	lockCommissionAmountAccountInfo, _ := accountCC.GenerateLockAccounts(internalUserAccountInfoForAddCryptocurrency.UserID, operationId)
	if lockCommissionAmountAccountInfo == nil {
		return nil, errors.Errorf("Lock commission amount AccountInfo can't by null or empty")
	}
	getCommissionArg := commissionCC.GetCommissionArg{
		PartnerCode: fromUser.PartnerCode, CustomerType: fromUser.CustomerType, ProcessType: transactionCC.DepositFunds_ProcessType,
	}
	commission, err := this.CommissionCC.GetCommission(getCommissionArg)
	if err != nil {
		return nil, err
	}

	// считаем комиссию за ввод средств
	commissionAmount, err := getCommissionAmountByAmountAndCommission(cryptocurrencyStore.Amount, commission.CommissionPercent)
	if err != nil {
		return nil, err
	}

	// снятие комиссии за ввод средств
	transactionCommissionRecords, err := this.depositCommission(lockCommissionAmountAccountInfo, internalUserAccountInfoForAddCryptocurrency, operationId, commissionAmount)
	result := append(transactionAmountRecords, transactionCommissionRecords...)
	if err != nil {
		return nil, err
	}

	return result, nil
}

// depositEthereumFundsTransaction создание/сохранение транзакции на вывод эфироподобной криптовалюты
func (this *TransactionManagerImpl) depositEthereumFundsTransaction(cryptocurrencyStore *cryptocurrencyCC.CryptocurrencyStore) ([]*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Infof("deposit ethereum funds transaction for cryptocurrencyStore %s", cryptocurrencyStore.ExternalTransactionID)

	if cryptocurrencyStore == nil {
		return nil, errors.Errorf("cryptocurrencyStore can't by null")
	}

	err := cryptocurrencyStore.Validate()
	if err != nil {
		return nil, logging.Get().Error(err.Error())
	}

	chaincode.GlobalLogger.Infof("search CodeCC CurrencyType %s Amount %s", cryptocurrencyStore.Amount.CurrencyType, cryptocurrencyStore.Amount.Amount)

	keysWithFilter := &requests.KeysWithFilter{}
	keysWithFilter.Keys = []string{cryptocurrencyStore.Amount.CurrencyType, cryptocurrencyStore.Amount.Amount}
	userKeyValueArray, err := this.CodeCC.GetKeyValueArrayByKeysWithFilter(keysWithFilter)
	if err != nil {
		return nil, errors.Errorf("can't find code by amount %s %s. Error: %s ", cryptocurrencyStore.Amount.CurrencyType, cryptocurrencyStore.Amount.Amount, err)
	}

	if userKeyValueArray == nil || len(userKeyValueArray.KeyValue) == 0 {
		return nil, errors.Errorf("can't find code by amount %s %s", cryptocurrencyStore.Amount.CurrencyType, cryptocurrencyStore.Amount.Amount)
	}

	if len(userKeyValueArray.KeyValue) > 1 {
		return nil, errors.Errorf("Found %d code by amount %s %s but expected 1", len(userKeyValueArray.KeyValue), cryptocurrencyStore.Amount.CurrencyType, cryptocurrencyStore.Amount.Amount)
	}

	internalUserAccountInfoForAddCryptocurrency, err := this.AccountCC.CheckAndLoadAccount(&accountCC.UserAccountInfo{UserID: userKeyValueArray.KeyValue[0].Value})
	if err != nil {
		return nil, errors.Errorf("failed to load information about internal user for deposit funds, due to %s", err)
	}

	chaincode.GlobalLogger.Infof("internalUserAccountInfoForAddCryptocurrency.AccountID: %s", internalUserAccountInfoForAddCryptocurrency.AccountID)
	chaincode.GlobalLogger.Infof("internalUserAccountInfoForAddCryptocurrency.WalletID: %s", internalUserAccountInfoForAddCryptocurrency.WalletID)
	chaincode.GlobalLogger.Infof("internalUserAccountInfoForAddCryptocurrency.UserID: %s", internalUserAccountInfoForAddCryptocurrency.UserID)

	operationId := fmt.Sprintf("%s.%s", internalUserAccountInfoForAddCryptocurrency, cryptocurrencyStore.ExternalTransactionID)

	transactionAmountRecords, err := this.depositAmount(internalUserAccountInfoForAddCryptocurrency, cryptocurrencyStore.Amount, operationId)
	if err != nil {
		return nil, err
	}

	fromUser, err := this.UserCC.GetUser(&requests.Id{Id: internalUserAccountInfoForAddCryptocurrency.UserID})
	if err != nil {
		return nil, err
	}

	// создаем аккаунт для блокировки средств на пользователя с комиссией
	lockCommissionAmountAccountInfo, _ := accountCC.GenerateLockAccounts(internalUserAccountInfoForAddCryptocurrency.UserID, operationId)
	if lockCommissionAmountAccountInfo == nil {
		return nil, errors.Errorf("Lock commission amount AccountInfo can't by null or empty")
	}
	getCommissionArg := commissionCC.GetCommissionArg{
		PartnerCode: fromUser.PartnerCode, CustomerType: fromUser.CustomerType, ProcessType: transactionCC.DepositFunds_ProcessType,
	}
	commission, err := this.CommissionCC.GetCommission(getCommissionArg)
	if err != nil {
		return nil, err
	}

	// считаем комиссию за ввод средств
	commissionAmount, err := getCommissionAmountByAmountAndCommission(cryptocurrencyStore.Amount, commission.CommissionPercent)
	if err != nil {
		return nil, err
	}

	// снятие комиссии за ввод средств
	transactionCommissionRecords, err := this.depositCommission(lockCommissionAmountAccountInfo, internalUserAccountInfoForAddCryptocurrency, operationId, commissionAmount)
	result := append(transactionAmountRecords, transactionCommissionRecords...)
	if err != nil {
		return nil, err
	}

	return result, nil
}

// комиссия за ввод средств
func (this *TransactionManagerImpl) depositCommission(lockCommissionAmountAccountInfo *accountCC.TxnAccountInfo, internalUserAccountInfoForAddCryptocurrency *accountCC.Account, operationId string, commissionAmount *nums.CurrencyNumber) ([]*transactionCC.TransactionRecord, error) {
	outTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			UserID: internalUserAccountInfoForAddCryptocurrency.UserID,
			TxnAccountInfo: &accountCC.TxnAccountInfo{
				WalletID:  internalUserAccountInfoForAddCryptocurrency.WalletID,
				AccountID: internalUserAccountInfoForAddCryptocurrency.AccountID,
			},
		},
		Amount:        commissionAmount.Plain(),
		OperationID:   operationId,
		OperationType: transactionCC.TransferCommissionToSystem,
		ProcessType:   transactionCC.DepositFunds_ProcessType,
		Timestamp:     transactionCC.Timestamp(this.GetTime()),
		InOutType:     transactionCC.OUTPUT,
	}

	inTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:         system_user.LOCK_USER,
			TxnAccountInfo: lockCommissionAmountAccountInfo,
		},
		Amount:        commissionAmount.Plain(),
		OperationID:   operationId,
		OperationType: transactionCC.TransferCommissionToSystem,
		ProcessType:   transactionCC.DepositFunds_ProcessType,
		Timestamp:     transactionCC.Timestamp(this.GetTime()),
		InOutType:     transactionCC.INPUT,
	}

	return this.SaveTransactionPair(outTransactionRecord, inTransactionRecord)
}

// сохранение пары транзакций
func (this *TransactionManagerImpl) SaveTransactionPair(outTransactionRecord *transactionCC.TransactionRecord, inTransactionRecord *transactionCC.TransactionRecord) ([]*transactionCC.TransactionRecord, error) {
	var transactionRecordList []*transactionCC.TransactionRecord

	// ключ для исходящей транзакции от пользователя - отправителя, для перевода
	// используется
	// - для сохранения в Value записи в леджере,
	// - как Key для сохранения в леджере
	outCompositeKey, err := this.CreateCompositeKeyByValidObject(outTransactionRecord)
	if err != nil {
		return nil, err
	}
	outTransactionRecord.Key = outCompositeKey

	// ключ для входящей транзакции к пользователю - получателю, для получения перевода
	// используется
	// - для сохранения в Value записи в леджере,
	// - как Key для сохранения в леджере
	inCompositeKey, err := this.CreateCompositeKeyByValidObject(inTransactionRecord)
	if err != nil {
		return nil, err
	}
	inTransactionRecord.Key = inCompositeKey

	// Связывание между собой Исходящей и Входящей транзакций
	inTransactionRecord.DependencyKey = outTransactionRecord.Key
	outTransactionRecord.DependencyKey = inTransactionRecord.Key

	_, err = this.Save(outTransactionRecord);
	if err != nil {
		return nil, err
	}

	// добавление outTransactionRecord в список всех успешных транзакций
	transactionRecordList = append(transactionRecordList, outTransactionRecord)

	_, err = this.Save(inTransactionRecord)
	if err != nil {
		return nil, err
	}

	// добавление inTransactionRecord в список всех успешных транзакций
	transactionRecordList = append(transactionRecordList, inTransactionRecord)

	// возвращаем список успешно проведенных транзакий
	return transactionRecordList, nil
}

// ввод средств
func (this *TransactionManagerImpl) depositAmount(internalUserAccountInfoForAddCryptocurrency *accountCC.Account, amount nums.PlainCurrencyNumber, operationId string) ([]*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Debugf("operationId: %s", operationId)

	isExistsTransactionByOperationId, err := isExistsTransactionByUserAccountInfoAndOperationId(internalUserAccountInfoForAddCryptocurrency.ToUserAccountInfo(), operationId, this)
	if err != nil {
		return nil, errors.Errorf("problem to execute function isExistsTransactionByOperationId, due to %e", err)
	}

	if isExistsTransactionByOperationId {
		return nil, errors.Errorf("transaction for account %s with operation_id %s already exists", internalUserAccountInfoForAddCryptocurrency.AccountID, operationId)
	}

	// load from account channel information about system user
	// responsible to deposit money
	beginUserInfo, err := this.AccountCC.CheckAndLoadAccount(&accountCC.UserAccountInfo{UserID: system_user.BEGIN_USER})
	if err != nil {
		return nil, errors.Errorf("failed to load information about system user to deposit funds, due to %s", err)
	}

	// we transferring money from system account
	beginUserAccountInfo := beginUserInfo.ToUserAccountInfo()

	// OUT
	outTransactionRecord := transactionCC.TransactionRecord{
		UserAccountInfo: beginUserAccountInfo,
		Amount:          amount,
		OperationID:     operationId,
		OperationType:   transactionCC.TransferAmountToUser,
		ProcessType:     transactionCC.DepositFunds_ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	// IN
	inTransactionRecord := transactionCC.TransactionRecord{
		UserAccountInfo: internalUserAccountInfoForAddCryptocurrency.ToUserAccountInfo(),
		Amount:          amount,
		OperationID:     operationId,
		OperationType:   transactionCC.TransferAmountToUser,
		ProcessType:     transactionCC.DepositFunds_ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.INPUT,
	}

	return this.SaveTransactionPair(&outTransactionRecord, &inTransactionRecord)
}

// OperationType:   TransferAmountToUser
// ProcessType:     DepositFunds_ProcessType
func isExistsTransactionByUserAccountInfoAndOperationId(userAccountInfo *accountCC.UserAccountInfo, operationId string, manager chaincode.IManager) (bool, error) {
	chaincode.GlobalLogger.Debugf("check is exists transaction by operationId %s", operationId)

	validObjects, err := manager.GetObjectListByKeysWithFilter(&transactionCC.TransactionRecord{}, []string{userAccountInfo.UserID, userAccountInfo.WalletID, userAccountInfo.AccountID}, operationId)
	if err != nil {
		return false, errors.Errorf("failed to check whenever locked account was spent, due to %s", err)
	}

	isExists := validObjects != nil && len(validObjects) != 0
	return isExists, nil
}

// OperationType:   TransferAmountToUser
// ProcessType:     DepositFunds_ProcessType
func isExistsTransactionByOperationId(operationId string, manager chaincode.IManager) (bool, error) {
	chaincode.GlobalLogger.Debugf("check is exists transaction by operationId %s", operationId)

	// add map for success transaction
	it, err := manager.GetStateByPartialCompositeKey(transactionCC.TransactionRecord{}.GetObjectType(), []string{operationId})
	if err != nil {
		return false, errors.Errorf("failed to check whenever locked account was spent, due to %s", err)
	}

	defer it.Close()
	if it.HasNext() {
		return true, nil
	}

	return false, nil
}

// получить сумму комиссии по сумме и проценту комиссии
func getCommissionAmountByAmountAndCommission(amount nums.PlainCurrencyNumber, commissionPercent float64) (*nums.CurrencyNumber, error) {
	number, err := amount.Number()
	if err != nil {
		return nil, err
	}

	currencyNumber, err := number.MulFloat(commissionPercent)
	if err != nil {
		return nil, err
	}

	return currencyNumber.DivSelfInt(100)
}

func (txnMngr *TransactionManagerImpl) isProcessCommissionTypeValidForProcessType(currentProcess transactionCC.ProcessType, conditionForCommission transactionCC.ProcessType) bool {
	chaincode.GlobalLogger.Infof("isProcessCommissionTypeValidForProcessType: %s - %s", currentProcess, conditionForCommission)

	if transactionCC.ExchangeWithBothCommission_ProcessType == currentProcess && conditionForCommission == transactionCC.TransferSenderCommission_ProcessType {
		return true
	} else if transactionCC.TransferSenderCommission_ProcessType == currentProcess && conditionForCommission == currentProcess {
		return true
	} else if transactionCC.TransferToMarket_ProcessType == currentProcess && conditionForCommission == transactionCC.TransferSenderCommission_ProcessType {
		return true
	} else {
		return false
	}
}

// проверка был ли уже перевод средств (разблокировка средств) по этой транзакции
func (this *TransactionManagerImpl) isFundsSpent(inputTransactionRecord *transactionCC.TransactionRecord) (error) {
	if inputTransactionRecord == nil {
		return logging.Get().Error("Failed. isFundsSpent. inputTransactionRecord can't be empty")
	}

	validObjects, err := this.GetObjectListByKeys(transactionCC.TransactionRecord{}, []string{inputTransactionRecord.UserAccountInfo.UserID, inputTransactionRecord.UserAccountInfo.WalletID, inputTransactionRecord.UserAccountInfo.AccountID, inputTransactionRecord.Amount.CurrencyType, inputTransactionRecord.OperationID, transactionCC.OUTPUT.String()})
	if err != nil {
		return err
	}

	if len(validObjects) == 0 {
		return nil
	}

	return logging.Get().Errorf("Failed. cannot spent same transaction twice, locked funds has been used %s", inputTransactionRecord.Key)
}

func (this *TransactionManagerImpl) FindTransactionByOperationIDAndOperationTypeAndUserAccountInfo(operationID string, operationType transactionCC.OperationType, userAccountInfo *accountCC.UserAccountInfo) (*transactionCC.TransactionList, *transactionCC.TransactionList, error) {
	if userAccountInfo == nil {
		errMsg := "userAccountInfo can't be nil, cannot process, FindTransactionByOperationIDAndOperationType request"

		return nil, nil, errors.Errorf(errMsg)
	}

	if &operationType == nil {
		errMsg := "operationType can't be nil, cannot process, FindTransactionByOperationIDAndOperationType request"

		return nil, nil, errors.Errorf(errMsg)
	}

	chaincode.GlobalLogger.Infof("FindTransactionByOperationIDAndOperationTypeAndUserAccountInfo("+
		"operationID: %s, operationType: %s, userAccountInfo.UserID: %s, userAccountInfo.WalletID: %s, userAccountInfo.AccountID: %s)",
		operationID, operationType.String(), userAccountInfo.UserID, userAccountInfo.WalletID, userAccountInfo.AccountID)

	if len(operationID) == 0 {
		errMsg := "operationID can't be empty, cannot process, FindTransactionByOperationIDAndOperationType request"

		return nil, nil, errors.Errorf(errMsg)
	}

	validObjects, err := this.GetObjectListByKeysWithFilter(&transactionCC.TransactionRecord{}, []string{userAccountInfo.UserID, userAccountInfo.WalletID, userAccountInfo.AccountID}, operationID, operationType.String())
	if err != nil {
		return nil, nil, err
	}

	if len(validObjects) == 0 {
		return nil, nil, errors.Errorf("Can not find transaction by operationID: %s, operationType: %s for user info %s", operationID, operationType, userAccountInfo.String())
	}

	var outTransactionList transactionCC.TransactionList
	var inTransactionList transactionCC.TransactionList
	for _, value := range validObjects {
		transactionRecord, err := transactionCC.Cast(value)
		if err != nil {
			return nil, nil, err
		}
		chaincode.GlobalLogger.Infof("found transaction with key %s", transactionRecord.Key)

		if transactionRecord.InOutType == transactionCC.INPUT {
			// add user transaction into list
			inTransactionList.Transactions = append(inTransactionList.Transactions, transactionRecord)
			chaincode.GlobalLogger.Infof("add new input transaction to list %s", len(outTransactionList.Transactions))

			// находим транзакцию откуда отправили средства
			dependencyTransaction, err := this.FindTransactionRecordByDependencyKey(transactionRecord.DependencyKey)
			if err != nil {
				return nil, nil, err
			}

			outTransactionList.Transactions = append(outTransactionList.Transactions, dependencyTransaction)
			chaincode.GlobalLogger.Infof("add new input transaction to list %s", len(outTransactionList.Transactions))
		} else if transactionRecord.InOutType == transactionCC.OUTPUT {
			// add user transaction into list
			outTransactionList.Transactions = append(outTransactionList.Transactions, transactionRecord)
			chaincode.GlobalLogger.Infof("add new output transaction to list %s", len(outTransactionList.Transactions))

			// находим транзакцию куда отправили средства
			dependencyTransaction, err := this.FindTransactionRecordByDependencyKey(transactionRecord.DependencyKey)
			if err != nil {
				return nil, nil, err
			}

			inTransactionList.Transactions = append(inTransactionList.Transactions, dependencyTransaction)
			chaincode.GlobalLogger.Infof("add new input transaction to list %s", len(inTransactionList.Transactions))
		}
	}

	chaincode.GlobalLogger.Infof("out transaction list size %d", len(outTransactionList.Transactions))
	chaincode.GlobalLogger.Infof("in transaction list size %d", len(inTransactionList.Transactions))

	if len(outTransactionList.Transactions) == 0 && len(inTransactionList.Transactions) == 0 {
		return nil, nil, errors.Errorf("Can not find transaction by operationID: %s, operationType: %s", operationID, operationType)
	}

	return &outTransactionList, &inTransactionList, nil
}

func (this *TransactionManagerImpl) FindTransactionRecordByDependencyKey(key string) (*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Infof("FindTransactionRecordByDependencyKey(key: %s)", key)

	if len(key) == 0 {
		return nil, errors.Errorf("an empty Key, cannot process, FindTransactionRecordByDependencyKey request")
	}

	validObject, err := this.GetObjectByCompositeKey(&transactionCC.TransactionRecord{}, key)
	if err != nil {
		return nil, err
	}

	if validObject == nil {
		return nil, errors.Errorf("Transaction not found by key: %s", key)
	}

	return transactionCC.Cast(validObject)
}
