package mngr

import (
	"github.com/pkg/errors"
	"github.com/shopspring/decimal"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/commissionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/model/system_user"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/userCC"
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
)

/***
Out user - пользователь, который отправляет средства;
In user - пользователь, которому отправляются средства;
*/

// LockFundsOnAccount lock funds for transfer from given accountID
// Функция блокировки средств для перевода с полученного идентификатора счета
// txnArgs - информация для блокировки средств на счёте пользователя
func (this *TransactionManagerImpl) LockFundsOnAccount(txnArgs *transactionCC.TxnFundsLockInput) (*transactionCC.TransactionList, error) {
	chaincode.GlobalLogger.Infof("LockFundsOnAccount(OutUser.UserID: %s, InUser.UserID: %s)", txnArgs.OutUser.UserID, txnArgs.InUser.UserID)

	// всегда возвращает строку содержащую id счета для блокировки
	lockCommissionAmountAccountInfo, lockAmountAccountInfo := accountCC.GenerateLockAccounts(txnArgs.OutUser.UserID, txnArgs.OperationID)

	// TODO удалить, т.к лишняя проверка
	if lockCommissionAmountAccountInfo == nil {
		return nil, errors.Errorf("Lock commission amount AccountInfo can't by null or empty")
	}

	// TODO удалить, т.к лишняя проверка
	if lockAmountAccountInfo == nil {
		return nil, errors.Errorf("Lock amount AccountInfo can't by null or empty")
	}

	// TODO перенести проверку в ValidateAmount
	// проверка, что тип валюты не нулевой
	if len(txnArgs.Amount.CurrencyType) == 0 {
		return nil, errors.Errorf("Currency type can't by null or empty")
	}

	// Проверка ошибки валидирования блокируемой суммы
	err := txnArgs.Amount.ValidateAmount()
	if err != nil {
		return nil, err
	}

	// Проверка, что переданное количество средств не пустое и не равняется нулю
	isAmountEmpty, err := txnArgs.Amount.IsAmountEmpty()
	if err != nil {
		return nil, err
	}
	if isAmountEmpty {
		return nil, chaincode.GlobalLogger.Error("amount can't be 0")
	}

	// Получаем исходящий счёт, значение которого получаем при вызове функции CheckAndLoadAccount в качестве параметра
	// передавая в эту функцию информацию о исходящем пользователе
	outAccount, err := this.AccountCC.CheckAndLoadAccount(txnArgs.OutUser)
	if err != nil {
		return nil, err
	}

	// Если значение не было получено, то выбрасываем ошибку
	if outAccount == nil {
		return nil, errors.Errorf("out user account info can't be null")
	}

	// Проводим валидацию счёта, проверяя, что все полученные поля заполнены
	err = outAccount.Validate()
	if err != nil {
		return nil, errors.Errorf("out user account {%s} is not valid {%s}", outAccount, err)
	}

	// Проверяем, что полученный счёт не заблокированный
	if outAccount.Status != accountCC.AccountStatus_ACTIVE {
		return nil, errors.Errorf("outUserAccountInfo expected Active, actual status %s", outAccount.Status)
	}

	// Проверяем, что строка с userId не равна нулю и то, что входящие и исходящие userID не равны между собой
	if (len(outAccount.UserID) != 0 && outAccount.UserID != txnArgs.OutUser.UserID) || (txnArgs.OutUser.TxnAccountInfo != nil &&
		((len(txnArgs.OutUser.TxnAccountInfo.AccountID) != 0 && outAccount.AccountID != txnArgs.OutUser.TxnAccountInfo.AccountID) ||
			(len(txnArgs.OutUser.TxnAccountInfo.WalletID) != 0 && (outAccount.WalletID != txnArgs.OutUser.TxnAccountInfo.WalletID)))) {
		return nil, errors.Errorf("loaded account %s need tobe equals user argument", outAccount.ToUserAccountInfo().String())
	}

	// Получаем входящий счёт, значение которого получаем при вызове функции из чейнкода AccountCC CheckAndLoadAccount
	// в качестве параметра передавая в эту функцию информацию о входящем пользователе

	// пример в txnArgs.InUser заполнено 2 поля AccountID и UserID
	// метод this.AccountCC.CheckAndLoadAccount(txnArgs.InUser) вернет полную информацию по счету (в которой есть WalletID)
	// вызов inAccount.ToUserAccountInfo() вернет поля AccountID и UserID + WalletID
	// далее для удобства мы приравняем txnArgs.InUser = inAccount.ToUserAccountInfo() чтобы заполнить все поля
	// в случае ошибки мы возвращаем ошибку с описанием почему не удалось подгрузить информацию по счету

	// *** Внимание поле UserID обязательно для заполнения -> поэтому мы сделали проверку выше
	inAccount, err := this.AccountCC.CheckAndLoadAccount(txnArgs.InUser)
	if err != nil {
		return nil, err
	}

	// Если значение не было получено, то выбрасываем ошибку
	if inAccount == nil {
		return nil, errors.Errorf("in user account info can't be null")
	}

	// Проводим валидацию счёта, проверяя, что все полученные поля заполнены
	err = inAccount.Validate()
	if err != nil {
		return nil, errors.Errorf("in user account {%s} is not valid {%s}", inAccount, err)
	}

	// Проверяем, что полученный счёт не заблокированный
	if inAccount.Status != accountCC.AccountStatus_ACTIVE {
		return nil, errors.Errorf("inAccount expected Active, actual status %s", inAccount.Status)
	}

	// Получаем всю информацию по входящему пользователю
	txnArgs.InUser = inAccount.ToUserAccountInfo()

	_, err = txnArgs.Amount.Number()
	if err != nil {
		return nil, err
	}

	// Get currency balance for user transferring funds from
	// Получить баланс по типу валюты для пользователя, который переводит средства
	chaincode.GlobalLogger.Debugf("Computing balance for account id %s, currency id %s", outAccount.AccountID, txnArgs.Amount.CurrencyType)
	currencyBalance, err := this.accountCurrencyBalance(outAccount.ToUserAccountInfo(), txnArgs.Amount.CurrencyType)
	if err != nil {
		return nil, errors.Errorf("failed to compute balance for account id %s and currency id %s, due to %s",
			outAccount.AccountID, txnArgs.Amount.CurrencyType, err)
	}

	// Получаем исходящего пользователя из чейнкода UserCC
	chaincode.GlobalLogger.Infof("userLoader(OutUser.UserID: %s)", txnArgs.OutUser.UserID)
	outUser, err := this.UserCC.GetUser(&requests.Id{Id: txnArgs.OutUser.UserID})
	if err != nil {
		return nil, err
	}
	chaincode.GlobalLogger.Infof("result: outUser.UserID: %s, outUser.PartnerCode: %s, outUser.CustomerType: %s, outUser.UserStatus: %s",
		outUser.UserID, outUser.PartnerCode, outUser.CustomerType, outUser.UserStatus)

	// Проверяем, что пользователь, который отправляет средства, не заблокирован
	if outUser.UserStatus == userCC.UserStatus_Blocked {
		return nil, errors.Errorf("To User is Blocked. UserID: %s; CustomerType: %s; PartnerCode: %s", outUser.UserID, outUser.CustomerType, outUser.PartnerCode)
	}

	// получение информации по комиссии (тип клиента "outUser.CustomerType") на вывод средств (тип комиссии "txnArgs.CommissionType") для пользователя и организации (id орагнизации "outUser.PartnerCode")
	getCommissionArg := commissionCC.GetCommissionArg{outUser.PartnerCode, outUser.CustomerType, txnArgs.ProcessType}
	commission, err := this.CommissionCC.GetCommission(getCommissionArg)
	if err != nil {
		return nil, err
	}

	// Получаем сумму комиссии передавая в функцию блокируемую сумму и процент комисии
	commissionAmount, err := getCommissionAmountByAmountAndCommission(txnArgs.Amount, commission.CommissionPercent)
	if err != nil {
		return nil, errors.Errorf("failed calc commissionAmount, due to %s", err)
	}

	// Если процесс не move, то пользователи должны быть разные, так как между своими счетами нельзя совершить операцию отличную от mover
	if txnArgs.ProcessType != transactionCC.Move_ProcessType && outAccount.UserID == inAccount.UserID {
		return nil, errors.Errorf("can't make transfer for self, use process type 'MOVE' instead of %s", txnArgs.ProcessType)
	}

	// Проверяем, что пользователь не совершает операцию move с одним и тем же счетом так как за move комиссия не снимается и пользователь может сделать бесконечное число таких запросов
	if txnArgs.ProcessType == transactionCC.Move_ProcessType && outAccount.AccountID == inAccount.AccountID {
		return nil, errors.Errorf("move process can't have same accountID %s", outAccount.AccountID)
	}

	// Если исходящий пользователь не равен входящему пользователю, то нельзя совершить операцию move
	if txnArgs.ProcessType == transactionCC.Move_ProcessType && outAccount.UserID != inAccount.UserID {
		return nil, errors.Errorf("can't make MOVE because out UserID %s != in UserID %s", outAccount.UserID, inAccount.UserID)
	}

	// Проверка, что у пользователя достаточно средств для оплаты комиссии + основная сумма которую пользователь запрашивает для вывода
	balance, err := currencyBalance.Balance.Number()
	if err != nil {
		return nil, err
	}

	totalAmount, err := txnArgs.Amount.Number()
	if err != nil {
		return nil, err
	}

	// If commission is on sender we need to lock commissions funds either
	// Если комиссия с отправителя, мы формируем totalAmount
	if txnArgs.ProcessType != transactionCC.TransferReceiverCommission_ProcessType && txnArgs.ProcessType != transactionCC.Move_ProcessType {
		totalAmount, err = totalAmount.Add(commissionAmount)
		if err != nil {
			return nil, err
		}
		// если комиссия с получателя
	} else if txnArgs.ProcessType == transactionCC.TransferReceiverCommission_ProcessType {
		if totalAmount == nil || totalAmount.Num == decimal.NewFromFloat(0) {
			return nil, errors.Errorf("totalAmount for transfer can't be 0")
		}
		if totalAmount.Num.LessThanOrEqual(commissionAmount.Num) {
			return nil, errors.Errorf("cannot lock %s funds for account id %s, due to commission %s from receiver equal or more than total amount %s",
				txnArgs.Amount.CurrencyType, outAccount.AccountID, commissionAmount.Num, totalAmount.Num)
		}
	}

	// Проверка достаточно ли средств на балансе пользователя
	err = this.isEnoughBalance(balance, totalAmount)
	if err != nil {
		return nil, errors.Errorf("cannot lock %s funds for account id %s, due to %s", txnArgs.Amount.CurrencyType, outAccount.AccountID, err)
	}

	// собираем список транзакция для ответа, в этом списке будут содержаться транзакции на in, out основных средств и in, out комиссии
	userTransactionsList := &transactionCC.TransactionList{}

	userTransactionsList.Transactions, err = this.lockFunds(lockAmountAccountInfo, txnArgs, outAccount)
	if err != nil {
		return nil, err
	}

	// If commission is on sender we need to lock commissions funds either
	// Если комиссия с отправителя, происходит блокировка суммы комиссии
	if txnArgs.ProcessType != transactionCC.TransferReceiverCommission_ProcessType && txnArgs.ProcessType != transactionCC.Move_ProcessType {
		transactionCommissionRecords, err := this.lockSenderCommission(lockCommissionAmountAccountInfo, txnArgs, commissionAmount, outAccount)
		userTransactionsList.Transactions = append(userTransactionsList.Transactions, transactionCommissionRecords...)
		if err != nil {
			return nil, err
		}
	}

	// Funds locked successfully
	// Сумма комисии заблокирована успешно
	return userTransactionsList, nil
}

// Функция блокировки средств от пользователя к другому пользователю
func (this *TransactionManagerImpl) lockFunds(lockAmountAccountInfo *accountCC.TxnAccountInfo, txnArgs *transactionCC.TxnFundsLockInput, outUserAccountInfo *accountCC.Account) ([]*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Info("create key for out user account")

	// Add lock amount transaction
	chaincode.GlobalLogger.Debugf("Locking funds %f on account id %s", txnArgs.Amount, outUserAccountInfo.AccountID)

	// сохранение исходящей транзакции от пользователя - отправителя на блокирующего пользователя, для блокировки средств на перевод
	outTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: outUserAccountInfo.ToUserAccountInfo(),
		Amount:          txnArgs.Amount,
		OperationID:     txnArgs.OperationID,
		OperationType:   transactionCC.LockAmount,
		ProcessType:     txnArgs.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	// сохранение входящей транзакции к блокирующему пользователю от пользователя - отправителя, для блокировки средств на перевод
	inTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:         system_user.LOCK_USER,
			TxnAccountInfo: lockAmountAccountInfo,
		},
		Amount:                     txnArgs.Amount,
		OperationID:                txnArgs.OperationID,
		OperationType:              transactionCC.LockAmount,
		ProcessType:                txnArgs.ProcessType,
		Timestamp:                  transactionCC.Timestamp(this.GetTime()),
		DestinationUserAccountInfo: txnArgs.InUser,
		InOutType:                  transactionCC.INPUT,
	}

	return this.SaveTransactionPair(outTransactionRecord, inTransactionRecord)
}

// Функция блокировки комиссии от пользователя к системе
func (this *TransactionManagerImpl) lockSenderCommission(lockCommissionAmountAccountInfo *accountCC.TxnAccountInfo, txnArgs *transactionCC.TxnFundsLockInput, commissionAmount *nums.CurrencyNumber, outUserAccountInfo *accountCC.Account) ([]*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Debugf("Locking commission funds %f on account id %s", commissionAmount, outUserAccountInfo.AccountID)

	outTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: outUserAccountInfo.ToUserAccountInfo(),
		Amount:          commissionAmount.Plain(),
		OperationID:     txnArgs.OperationID,
		OperationType:   transactionCC.LockCommission,
		ProcessType:     txnArgs.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	inTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:         system_user.LOCK_USER,
			TxnAccountInfo: lockCommissionAmountAccountInfo,
		},
		Amount:        commissionAmount.Plain(),
		OperationID:   txnArgs.OperationID,
		OperationType: transactionCC.LockCommission,
		ProcessType:   txnArgs.ProcessType,
		Timestamp:     transactionCC.Timestamp(this.GetTime()),
		InOutType:     transactionCC.INPUT,
	}

	return this.SaveTransactionPair(outTransactionRecord, inTransactionRecord)
}
