package mngr

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/userCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/system_user"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/model/commissionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
)

// Блокировка средств на вывод из системы на END_USER
func (this *TransactionManagerImpl) LockOutFundsOnAccount(txnArgs *transactionCC.OutTxnFundsLockInput) (*transactionCC.TransactionList, error) {
	userTransactionsList := &transactionCC.TransactionList{}

	if len(txnArgs.Amount.CurrencyType) == 0 {
		return userTransactionsList, errors.Errorf("Currency type can't by null or empty")
	}

	if txnArgs.ProcessType == transactionCC.WithdrawalFunds_ProcessType {
		return userTransactionsList, errors.Errorf("ProcessType for lock out funds on account need to be equals 'OUT_FOUNDS'")
	}

	// ASSUMPTION: All user ids are present, otherwise chaincode invocation will results with error
	if len(txnArgs.OutUser.UserID) == 0 {
		return userTransactionsList, errors.Errorf("missing from user id, operation id %s", txnArgs.OperationID)
	}

	// Verify to user information
	fromUserAccountInfo, err := this.AccountCC.CheckAndLoadAccount(txnArgs.OutUser)
	if err != nil {
		return nil, err
	}

	_, err = txnArgs.Amount.Number()
	if err != nil {
		return nil, err
	}

	// Get currency balance for user transferring funds from
	chaincode.GlobalLogger.Debugf("Computing balance for account id %s, currency id %s", fromUserAccountInfo.AccountID, txnArgs.Amount.CurrencyType)
	currencyBalance, err := this.accountCurrencyBalance(fromUserAccountInfo.ToUserAccountInfo(), txnArgs.Amount.CurrencyType)
	if err != nil {
		return nil, errors.Errorf("failed to compute balance for account id %s and currency id %s, due to %s",
			fromUserAccountInfo.AccountID, txnArgs.Amount.CurrencyType, err)
	}

	chaincode.GlobalLogger.Infof("userLoader(OutUser.UserID: %s)", txnArgs.OutUser.UserID)

	fromUser, err := this.UserCC.GetUser(&requests.Id{Id: txnArgs.OutUser.UserID})
	if err != nil {
		return nil, err
	}

	chaincode.GlobalLogger.Infof("result: fromUser.UserID: %s, fromUser.PartnerCode: %s, fromUser.CustomerType: %s, fromUser.UserStatus: %s",
		fromUser.UserID, fromUser.PartnerCode, fromUser.CustomerType, fromUser.UserStatus)

	if fromUser.UserStatus == userCC.UserStatus_Blocked {
		return userTransactionsList, errors.Errorf("To User is Blocked. UserID: %s; CustomerType: %s; PartnerCode: %s", fromUser.UserID, fromUser.CustomerType, fromUser.PartnerCode)
	}

	// подсчет комиссии за вывод средств
	getCommissionArg := commissionCC.GetCommissionArg{
		PartnerCode: fromUser.PartnerCode, CustomerType: fromUser.CustomerType, ProcessType: txnArgs.ProcessType,
	}
	commission, err := this.CommissionCC.GetCommission(getCommissionArg)
	if err != nil {
		return nil, err
	}

	commissionAmount, err := getCommissionAmountByAmountAndCommission(txnArgs.Amount, commission.CommissionPercent)
	if err != nil {
		return nil, err
	}
	// Commission is on sender hence need to check whenever he
	// has enough funds including commission,
	// is commissionType != SenderCommission => commission will be zero
	balance, err := currencyBalance.Balance.Number()
	if err != nil {
		return nil, err
	}

	amount, err := txnArgs.Amount.Number()
	if err != nil {
		return nil, err
	}

	totalAmount, err := amount.Add(commissionAmount)
	if err != nil {
		return nil, err
	}

	err = this.isEnoughBalance(balance, totalAmount)

	if err != nil {
		return nil, errors.Errorf("cannot lock %s funds for account id %s, due to %s", txnArgs.Amount.CurrencyType, fromUserAccountInfo.AccountID, err)
	}

	chaincode.GlobalLogger.Infof("LockOutFundsOnAccount(OutUser.UserID: %s)", txnArgs.OutUser.UserID)
	lockCommissionAmountAccountInfo, lockAmountAccountInfo := accountCC.GenerateLockAccounts(txnArgs.OutUser.UserID, txnArgs.OperationID)

	transactionRecords, err := this.lockOutAmount(lockAmountAccountInfo, fromUserAccountInfo, txnArgs)
	if err != nil {
		return nil, err
	}

	lockTransactionCommissionRecords, err := this.lockOutCommission(lockCommissionAmountAccountInfo, txnArgs, commissionAmount, fromUserAccountInfo)
	transactionRecords = append(transactionRecords, lockTransactionCommissionRecords...)
	if err != nil {
		return nil, err
	}

	userTransactionsList.Transactions = transactionRecords
	// Funds locked successfully
	return userTransactionsList, nil
}


// блокировка средств на вывод из системы
func (this *TransactionManagerImpl) lockOutAmount(lockAmountAccountInfo *accountCC.TxnAccountInfo, fromUserAccountInfo *accountCC.Account, txnArgs *transactionCC.OutTxnFundsLockInput) ([]*transactionCC.TransactionRecord, error) {
	isAmountEmpty, err := txnArgs.Amount.IsAmountEmpty()
	if err != nil {
		return nil, err
	}
	if isAmountEmpty {
		return nil, chaincode.GlobalLogger.Error("amount can't be 0")
	}

	endUserAccountInfo, err := this.AccountCC.CheckAndLoadAccount(&accountCC.UserAccountInfo{UserID: system_user.END_USER})
	if err != nil {
		return nil, err
	}

	// Add lock amount transaction
	chaincode.GlobalLogger.Debugf("Locking funds %f on account id %s", txnArgs.Amount, fromUserAccountInfo.AccountID)

	outTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: fromUserAccountInfo.ToUserAccountInfo(),
		Amount:          txnArgs.Amount,
		OperationID:     txnArgs.OperationID,
		OperationType:   transactionCC.LockAmount,
		ProcessType:     txnArgs.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	// Create reciprocal transaction with inverse keys and system user
	inTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:         system_user.LOCK_USER,
			TxnAccountInfo: lockAmountAccountInfo,
		},
		Amount:                     txnArgs.Amount,
		OperationID:                txnArgs.OperationID,
		OperationType:              transactionCC.LockAmount,
		ProcessType:                txnArgs.ProcessType,
		Timestamp:                  transactionCC.Timestamp(this.GetTime()),
		DestinationUserAccountInfo: endUserAccountInfo.ToUserAccountInfo(),
		InOutType:                  transactionCC.INPUT,
	}

	return this.SaveTransactionPair(outTransactionRecord, inTransactionRecord)
}

// блокировка комиссии за вывод средств
func (this *TransactionManagerImpl) lockOutCommission(lockCommissionAmountAccountInfo *accountCC.TxnAccountInfo,
	txnArgs *transactionCC.OutTxnFundsLockInput, commissionAmount *nums.CurrencyNumber, fromUserAccountInfo *accountCC.Account) ([]*transactionCC.TransactionRecord, error) {

	chaincode.GlobalLogger.Debugf("Locking commission funds %f on account id %s", commissionAmount, fromUserAccountInfo.AccountID)
	outTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: fromUserAccountInfo.ToUserAccountInfo(),
		Amount:          commissionAmount.Plain(),
		OperationID:     txnArgs.OperationID,
		OperationType:   transactionCC.LockCommission,
		ProcessType:     txnArgs.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	// The reciprocal commission lock record
	inTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:         system_user.LOCK_USER,
			TxnAccountInfo: lockCommissionAmountAccountInfo,
		},
		Amount:        commissionAmount.Plain(),
		OperationID:   txnArgs.OperationID,
		OperationType: transactionCC.LockCommission,
		ProcessType:   txnArgs.ProcessType,
		Timestamp:     transactionCC.Timestamp(this.GetTime()),
		InOutType:     transactionCC.INPUT,
	}

	return this.SaveTransactionPair(inTransactionRecord, outTransactionRecord)
}
