package mngr

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/userCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/system_user"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/model/commissionCC"
)

// Блокировка средств на перевод средств на биржу на MARKET_USER
func (this *TransactionManagerImpl) LockTransferToMarket(txnArgs *transactionCC.OutTxnFundsLockInput) error {
	chaincode.GlobalLogger.Infof("LockTransferToMarket(OutUser.UserID: %s)", txnArgs.OutUser.UserID)

	lockCommissionAmountAccountInfo, lockAmountAccountInfo := accountCC.GenerateLockAccounts(txnArgs.OutUser.UserID, txnArgs.OperationID)

	if len(txnArgs.Amount.CurrencyType) == 0 {
		return errors.Errorf("Currency type can't by null or empty")
	}

	// ASSUMPTION: All user ids are present, otherwise chaincode invocation will results with error
	if len(txnArgs.OutUser.UserID) == 0 {
		return errors.Errorf("missing from user id, operation id %s", txnArgs.OperationID)
	}

	// Verify to user information
	fromUserAccountInfo, err := this.AccountCC.CheckAndLoadAccount(txnArgs.OutUser)
	if err != nil {
		return err
	}

	err = txnArgs.Amount.ValidateAmount()
	if err != nil {
		return err
	}

	err = txnArgs.Amount.Validate()
	if err != nil {
		return err
	}

	isAmountEmpty, err := txnArgs.Amount.IsAmountEmpty()
	if err != nil {
		return err
	}
	if isAmountEmpty {
		return chaincode.GlobalLogger.Error("amount can't be 0")
	}

	// Get currency balance for user transferring funds from
	chaincode.GlobalLogger.Debugf("Computing balance for account id %s, currency id %s", fromUserAccountInfo.AccountID, txnArgs.Amount.CurrencyType)
	currencyBalance, err := this.accountCurrencyBalance(fromUserAccountInfo.ToUserAccountInfo(), txnArgs.Amount.CurrencyType)
	if err != nil {
		return errors.Errorf("failed to compute balance for account id %s and currency id %s, due to %s",
			fromUserAccountInfo.AccountID, txnArgs.Amount.CurrencyType, err)
	}

	chaincode.GlobalLogger.Infof("userLoader(OutUser.UserID: %s)", txnArgs.OutUser.UserID)
	fromUser, err := this.UserCC.GetUser(&requests.Id{Id: txnArgs.OutUser.UserID})
	if err != nil {
		return err
	}

	chaincode.GlobalLogger.Infof("result: fromUser.UserID: %s, fromUser.PartnerCode: %s, fromUser.CustomerType: %s, fromUser.UserStatus: %s", fromUser.UserID, fromUser.PartnerCode, fromUser.CustomerType, fromUser.UserStatus)
	if fromUser.UserStatus == userCC.UserStatus_Blocked {
		return errors.Errorf("To User is Blocked. UserID: %s; CustomerType: %s; PartnerCode: %s", fromUser.UserID, fromUser.CustomerType, fromUser.PartnerCode)
	}

	var commissionPercent float64
	if txnArgs.ProcessType != transactionCC.TransferReceiverCommission_ProcessType {
		// Check whenever we have enough money to transfer and pay commission
		getCommissionArg := commissionCC.GetCommissionArg{
			PartnerCode: fromUser.PartnerCode, CustomerType: fromUser.CustomerType, ProcessType: txnArgs.ProcessType,
		}
		commission, err := this.CommissionCC.GetCommission(getCommissionArg)
		if err != nil {
			return err
		}
		commissionPercent = commission.CommissionPercent
	}

	commissionAmount, err := getCommissionAmountByAmountAndCommission(txnArgs.Amount, commissionPercent)
	if err != nil {
		return err
	}

	// Commission is on sender hence need to check whenever he
	// has enough funds including commission,
	// is commissionType != model2.SenderCommission => commission will be zero
	balance, err := currencyBalance.Balance.Number()
	if err != nil {
		return err
	}

	amount, err := txnArgs.Amount.Number()
	if err != nil {
		return err
	}

	totalAmount, err := amount.Add(commissionAmount)
	if err != nil {
		return err
	}

	err = this.isEnoughBalance(balance, totalAmount)
	if err != nil {
		return errors.Errorf("cannot lock %s funds for account id %s, due to %s", txnArgs.Amount.CurrencyType, fromUserAccountInfo.AccountID, err)
	}

	// transaction record in a form:

	/*
		####################################################################################################################
			out.CurrencyType.FromUserAccountID,...,FromUserID,in.CurrencyType.LockAccountUserID (LockAccountUserID == system account id)
		####################################################################################################################
		 	need to add reciprocal transaction record, e.g.:
		 	in.CurrencyType.FromUserAccountID,...,SystemUserID,out.CurrencyType.FromUserAccountID (LockAccountUserID == system account id)
		####################################################################################################################
	*/

	// Add lock amount transaction
	chaincode.GlobalLogger.Infof("Locking funds %f on account id %s", txnArgs.Amount, fromUserAccountInfo.AccountID)

	// Copy transaction data
	_, err = this.Save(&transactionCC.TransactionRecord{
		UserAccountInfo: fromUserAccountInfo.ToUserAccountInfo(),
		Amount:          txnArgs.Amount,
		OperationID:     txnArgs.OperationID,
		OperationType:   transactionCC.LockAmount,
		ProcessType:     txnArgs.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	})
	if err != nil {
		return err
	}

	marketUserAccountInfo, err := this.AccountCC.CheckAndLoadAccount(&accountCC.UserAccountInfo{UserID: system_user.MARKET_OUT_USER})
	if err != nil {
		return err
	}

	// Create reciprocal transaction with inverse keys and system user
	_, err = this.Save(&transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:         system_user.LOCK_USER,
			TxnAccountInfo: lockAmountAccountInfo,
		},
		Amount:                     txnArgs.Amount,
		OperationID:                txnArgs.OperationID,
		OperationType:              transactionCC.LockAmount,
		ProcessType:                txnArgs.ProcessType,
		Timestamp:                  transactionCC.Timestamp(this.GetTime()),
		DestinationUserAccountInfo: marketUserAccountInfo.ToUserAccountInfo(),
		InOutType:                  transactionCC.INPUT,
	})
	if err != nil {
		return err
	}

	// If commission is on sender we need to lock commissions funds either
	if txnArgs.ProcessType != transactionCC.TransferReceiverCommission_ProcessType {
		// Copy transaction data
		chaincode.GlobalLogger.Debugf("Locking commission funds %f on account id %s", commissionAmount, fromUserAccountInfo.AccountID)
		_, err = this.Save(&transactionCC.TransactionRecord{
			UserAccountInfo: fromUserAccountInfo.ToUserAccountInfo(),
			Amount:          commissionAmount.Plain(),
			OperationID:     txnArgs.OperationID,
			OperationType:   transactionCC.LockCommission,
			ProcessType:     txnArgs.ProcessType,
			Timestamp:       transactionCC.Timestamp(this.GetTime()),
			InOutType:       transactionCC.OUTPUT,
		})
		if err != nil {
			return err
		}

		// The reciprocal commission lock record
		_, err = this.Save(&transactionCC.TransactionRecord{
			UserAccountInfo: &accountCC.UserAccountInfo{
				// Keep default values for time being, has to be replaced with
				// call to system chaincode to obtain system user and account information
				UserID:         system_user.LOCK_USER,
				TxnAccountInfo: lockCommissionAmountAccountInfo,
			},
			Amount:        commissionAmount.Plain(),
			OperationID:   txnArgs.OperationID,
			OperationType: transactionCC.LockCommission,
			ProcessType:   txnArgs.ProcessType,
			Timestamp:     transactionCC.Timestamp(this.GetTime()),
			InOutType:     transactionCC.INPUT,
		})
		if err != nil {
			return err
		}
	}

	// Funds locked successfully
	return nil
}
