package mngr

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/commissionCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/system_user"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/userCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
)

// TransferFunds finalize transferring funds from locked account to the end user
func (this *TransactionManagerImpl) transferFunds(lockTransactionRecordInfo *transactionCC.LockTransactionRecordInfo) (*transactionCC.TransactionList, error) {
	// START validation argument
	if lockTransactionRecordInfo == nil {
		errMsg := "Lock transaction record info can't be nil"
		return nil, errors.Errorf(errMsg)
	}

	err := lockTransactionRecordInfo.Validate()
	if err != nil {
		return nil, err
	}

	chaincode.GlobalLogger.Infof("load systemCommissionUser %s", system_user.SYSTEM_COMMISSION_USER)
	systemCommissionUserInfo, err := this.AccountCC.CheckAndLoadAccount(&accountCC.UserAccountInfo{UserID: system_user.SYSTEM_COMMISSION_USER})
	if err != nil {
		return nil, errors.Errorf("failed to load information about system user to deposit funds, due to %s", err)
	}

	inLockedAmountTransactionRecord := lockTransactionRecordInfo.InLockAmountTransactionRecord
	outLockedAmountTransactionRecord := lockTransactionRecordInfo.OutLockAmountTransactionRecord
	inLockedCommissionTransactionRecord := lockTransactionRecordInfo.InLockCommissionTransactionRecord
	operationID := outLockedAmountTransactionRecord.OperationID
	chaincode.GlobalLogger.Infof("TransferFunds. OperationID: %s", operationID)

	if systemCommissionUserInfo == nil {
		msg := "System commission user info can't be nil"
		chaincode.GlobalLogger.Error(msg)
		return nil, errors.Errorf(msg)
	}

	inUser, err := this.UserCC.GetUser(&requests.Id{Id: inLockedAmountTransactionRecord.DestinationUserAccountInfo.UserID})
	if err != nil {
		return nil, err
	}

	if inUser.UserStatus == userCC.UserStatus_Blocked {
		return nil, errors.Errorf("To user is Blocked. UserID: %s; CustomerType: %s; PartnerCode: %s", inUser.UserID, inUser.CustomerType, inUser.PartnerCode)
	}

	// Verify to user information
	_, err = this.AccountCC.CheckAndLoadAccount(inLockedAmountTransactionRecord.DestinationUserAccountInfo)
	if err != nil {
		return nil, err
	}

	// Check whenever funds on locked account was spent
	err = this.isFundsSpent(inLockedAmountTransactionRecord)
	if err != nil {
		return nil, err
	}

	// if commission is on sender need to check whenever commission was locked on temp account
	if inLockedAmountTransactionRecord.ProcessType != transactionCC.TransferReceiverCommission_ProcessType  && inLockedAmountTransactionRecord.ProcessType !=  transactionCC.Move_ProcessType {
		// Check whenever funds on locked account for commission was spent
		err := this.isFundsSpent(inLockedCommissionTransactionRecord)
		if err != nil {
			return nil, err
		}
	}

	userTransactionsList := &transactionCC.TransactionList{}

	// перевод средств
	userTransactionsList.Transactions, err = this.transfer(inLockedAmountTransactionRecord)
	if err != nil {
		return userTransactionsList, err
	}

	// перевод комиссии - комиссия с получателя
	// && inLockedAmountTransactionRecord.ProcessType !=  transactionCC.Move_ProcessType
	if inLockedAmountTransactionRecord.ProcessType == transactionCC.TransferReceiverCommission_ProcessType {
		inLockedCommissionTransactionRecord, lockTransactionCommissionRecords, err := this.lockReceiverCommission(systemCommissionUserInfo, inLockedAmountTransactionRecord)
		userTransactionsList.Transactions = append(userTransactionsList.Transactions, lockTransactionCommissionRecords...)
		if err != nil {
			return userTransactionsList, err
		}

		transactionCommissionRecords, err := this.transferCommission(inLockedCommissionTransactionRecord, systemCommissionUserInfo)
		userTransactionsList.Transactions = append(userTransactionsList.Transactions, transactionCommissionRecords...)
		if err != nil {
			return nil, err
		}
	}

	// перевод комиссии - комиссия с отправителя
	if inLockedAmountTransactionRecord.ProcessType != transactionCC.TransferReceiverCommission_ProcessType && inLockedAmountTransactionRecord.ProcessType !=  transactionCC.Move_ProcessType {
		transactionCommissionRecords, err := this.transferCommission(inLockedCommissionTransactionRecord, systemCommissionUserInfo)
		userTransactionsList.Transactions = append(userTransactionsList.Transactions, transactionCommissionRecords...)
		if err != nil {
			return nil, err
		}
	}

	return userTransactionsList, nil
}

// перевод средств от пользователя к другому пользователю
func (this *TransactionManagerImpl) transfer(inLockedAmountTransactionRecord *transactionCC.TransactionRecord) ([]*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Infof("TransferFunds. Transfer lockedAmountTransactionRecord(%s) to user %s", inLockedAmountTransactionRecord.Key, inLockedAmountTransactionRecord.DestinationUserAccountInfo.UserID)
	chaincode.GlobalLogger.Debugf("Transferring funds from locked account %s to account id %s", inLockedAmountTransactionRecord.UserAccountInfo.AccountID, inLockedAmountTransactionRecord.DestinationUserAccountInfo.AccountID)

	// сохранение исходящей транзакции от пользователя - отправителя, для перевода
	outTransferringFundsFromLockedAccount := &transactionCC.TransactionRecord{
		UserAccountInfo: inLockedAmountTransactionRecord.UserAccountInfo,
		Amount:          inLockedAmountTransactionRecord.Amount,
		OperationID:     inLockedAmountTransactionRecord.OperationID,
		OperationType:   transactionCC.TransferAmountToUser,
		ProcessType:     inLockedAmountTransactionRecord.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	// сохранение входещей транзакции к пользователю - получателю, для получения перевода
	inTransferringFundsFromLockedAccount := &transactionCC.TransactionRecord{
		UserAccountInfo: inLockedAmountTransactionRecord.DestinationUserAccountInfo,
		Amount:          inLockedAmountTransactionRecord.Amount,
		OperationID:     inLockedAmountTransactionRecord.OperationID,
		OperationType:   transactionCC.TransferAmountToUser,
		ProcessType:     inLockedAmountTransactionRecord.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.INPUT,
	}

	return this.SaveTransactionPair(outTransferringFundsFromLockedAccount, inTransferringFundsFromLockedAccount)
}

// блокировка средств от пользователя к другому пользователю
func (this *TransactionManagerImpl) lockReceiverCommission(systemCommissionUserInfo *accountCC.Account, inLockedAmountTransactionRecord *transactionCC.TransactionRecord)  (*transactionCC.TransactionRecord, []*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Info("create key for out user account")

	lockCommissionAmountAccountInfo, _ := accountCC.GenerateLockAccounts(inLockedAmountTransactionRecord.DestinationUserAccountInfo.UserID, inLockedAmountTransactionRecord.OperationID)

	outUser, err := this.UserCC.GetUser(&requests.Id{Id: inLockedAmountTransactionRecord.DestinationUserAccountInfo.UserID})
	if err != nil {
		return nil, nil, err
	}
	getCommissionArg := commissionCC.GetCommissionArg{outUser.PartnerCode, outUser.CustomerType, inLockedAmountTransactionRecord.ProcessType}
	commission, err := this.CommissionCC.GetCommission(getCommissionArg)
	if err != nil {
		return nil, nil, err
	}

	commissionAmount, err := getCommissionAmountByAmountAndCommission(inLockedAmountTransactionRecord.Amount, commission.CommissionPercent)
	if err != nil {
		return nil, nil, errors.Errorf("failed calc commissionAmount, due to %s", err)
	}

	chaincode.GlobalLogger.Debugf("Locking commission funds %f on account id %s", commissionAmount, inLockedAmountTransactionRecord.DestinationUserAccountInfo.AccountID)
	outTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: inLockedAmountTransactionRecord.DestinationUserAccountInfo,
		Amount:          commissionAmount.Plain(),
		OperationID:     inLockedAmountTransactionRecord.OperationID,
		OperationType:   transactionCC.LockCommission,
		ProcessType:     inLockedAmountTransactionRecord.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	inTransactionRecord := &transactionCC.TransactionRecord{
		UserAccountInfo: &accountCC.UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:         system_user.LOCK_USER,
			TxnAccountInfo: lockCommissionAmountAccountInfo,
		},
		Amount:        commissionAmount.Plain(),
		OperationID:   inLockedAmountTransactionRecord.OperationID,
		OperationType: transactionCC.LockCommission,
		ProcessType:   inLockedAmountTransactionRecord.ProcessType,
		Timestamp:     transactionCC.Timestamp(this.GetTime()),
		InOutType:     transactionCC.INPUT,
	}

	transactionRecords, err := this.SaveTransactionPair(outTransactionRecord, inTransactionRecord)
	return inTransactionRecord, transactionRecords, err
}

// перевод комиссии от пользователя к системному пользователю
func (this *TransactionManagerImpl) transferCommission(inLockedCommissionTransactionRecord *transactionCC.TransactionRecord, systemCommissionUserInfo *accountCC.Account) ([]*transactionCC.TransactionRecord, error) {
	chaincode.GlobalLogger.Infof("TransferFunds. Transfer lockedCommissionTransactionRecord(%s) to user %s", inLockedCommissionTransactionRecord.Key, systemCommissionUserInfo.UserID)
	chaincode.GlobalLogger.Debugf("Transferring commission from locked account %s to the system account", inLockedCommissionTransactionRecord.UserAccountInfo.AccountID)

	// сохранение исходящей транзакции от пользователя для отправки комиссии
	outTransferringCommissionFromLockedAccount := &transactionCC.TransactionRecord{
		UserAccountInfo: inLockedCommissionTransactionRecord.UserAccountInfo,
		Amount:          inLockedCommissionTransactionRecord.Amount,
		OperationID:     inLockedCommissionTransactionRecord.OperationID,
		OperationType:   transactionCC.TransferCommissionToSystem,
		ProcessType:     inLockedCommissionTransactionRecord.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	// сохранение входещей транзакции к системному пользователю для получения комиссии
	inTransferringCommissionFromLockedAccount := &transactionCC.TransactionRecord{
		UserAccountInfo: systemCommissionUserInfo.ToUserAccountInfo(),
		Amount:          inLockedCommissionTransactionRecord.Amount,
		OperationID:     inLockedCommissionTransactionRecord.OperationID,
		OperationType:   transactionCC.TransferCommissionToSystem,
		ProcessType:     inLockedCommissionTransactionRecord.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.INPUT,
	}

	return this.SaveTransactionPair(outTransferringCommissionFromLockedAccount, inTransferringCommissionFromLockedAccount)
}