package mngr

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
)

// TransferFunds finalize transferring funds from locked account to the end user
func (this *TransactionManagerImpl) unlockTransferFunds(lockTransactionRecordInfo *transactionCC.LockTransactionRecordInfo) (*transactionCC.TransactionList, error) {
	// START validation argument
	if lockTransactionRecordInfo == nil {
		errMsg := "Lock transaction record info can't be nil"
		return nil, errors.Errorf(errMsg)
	}

	err := lockTransactionRecordInfo.Validate()
	if err != nil {
		return nil, err
	}

	// unlock amount to sender user
	inLockedAmountTransactionRecord := lockTransactionRecordInfo.InLockAmountTransactionRecord // на ком сейчас находятся заблокированные средства
	outLockAmountTransactionRecord := lockTransactionRecordInfo.OutLockAmountTransactionRecord // кто изначально хочет отправить средства другому пользователю

	// unlock commission to sender user
	inLockCommissionTransactionRecord := lockTransactionRecordInfo.InLockCommissionTransactionRecord // на ком сейчас находятся заблокированные средства
	outLockCommissionTransactionRecord := lockTransactionRecordInfo.OutLockCommissionTransactionRecord // кто изначально хочет отправить средства другому пользователю

	// from lock user - inLockedAmountTransactionRecord to outLockAmountTransactionRecord

	// формирования ключей для хранения в ledger

	// Copy transaction data
	chaincode.GlobalLogger.Debugf("Transferring funds from locked account %s to account id %s", inLockedAmountTransactionRecord.UserAccountInfo.AccountID, inLockedAmountTransactionRecord.DestinationUserAccountInfo.AccountID)
	outTransferringFundsFromLockedAccount := &transactionCC.TransactionRecord{
		UserAccountInfo: inLockedAmountTransactionRecord.UserAccountInfo,
		Amount:          inLockedAmountTransactionRecord.Amount,
		OperationID:     inLockedAmountTransactionRecord.OperationID,
		OperationType:   transactionCC.UnlockAmount,
		ProcessType:     inLockedAmountTransactionRecord.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.OUTPUT,
	}

	inTransferringFundsFromLockedAccount := &transactionCC.TransactionRecord{
		UserAccountInfo: outLockAmountTransactionRecord.UserAccountInfo,
		Amount:          outLockAmountTransactionRecord.Amount,
		OperationID:     outLockAmountTransactionRecord.OperationID,
		OperationType:   transactionCC.UnlockAmount,
		ProcessType:     outLockAmountTransactionRecord.ProcessType,
		Timestamp:       transactionCC.Timestamp(this.GetTime()),
		InOutType:       transactionCC.INPUT,
	}

	userTransactionsList := &transactionCC.TransactionList{}

	transactionRecords, err := this.SaveTransactionPair(outTransferringFundsFromLockedAccount, inTransferringFundsFromLockedAccount)
	if err != nil {
		return nil, err
	}
	userTransactionsList.Transactions = append(userTransactionsList.Transactions, transactionRecords...)

	if outLockCommissionTransactionRecord != nil && inLockCommissionTransactionRecord != nil {

		chaincode.GlobalLogger.Debugf("Transferring commission from locked account %s to the system account", inLockCommissionTransactionRecord.UserAccountInfo.AccountID)
		outTransferringCommissionFromLockedAccount := &transactionCC.TransactionRecord{
			UserAccountInfo: inLockCommissionTransactionRecord.UserAccountInfo,
			Amount:          inLockCommissionTransactionRecord.Amount,
			OperationID:     inLockCommissionTransactionRecord.OperationID,
			OperationType:   transactionCC.UnlockCommission,
			ProcessType:     inLockCommissionTransactionRecord.ProcessType,
			Timestamp:       transactionCC.Timestamp(this.GetTime()),
			InOutType:       transactionCC.OUTPUT,
		}

		inTransferringCommissionFromLockedAccount := &transactionCC.TransactionRecord{
			UserAccountInfo: outLockCommissionTransactionRecord.UserAccountInfo,
			Amount:          outLockCommissionTransactionRecord.Amount,
			OperationID:     outLockCommissionTransactionRecord.OperationID,
			OperationType:   transactionCC.UnlockCommission,
			ProcessType:     outLockCommissionTransactionRecord.ProcessType,
			Timestamp:       transactionCC.Timestamp(this.GetTime()),
			InOutType:       transactionCC.INPUT,
		}

		transactionPair, err := this.SaveTransactionPair(outTransferringCommissionFromLockedAccount, inTransferringCommissionFromLockedAccount)
		if err != nil {
			return nil, err
		}
		userTransactionsList.Transactions = append(userTransactionsList.Transactions, transactionPair...)

		return userTransactionsList, nil
	}

	return userTransactionsList, nil
}
