package chaincode

import (
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"encoding/json"
	"fmt"
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
)

var (
	GlobalLogger = logging.Get()
)

var CommonChaincodeActionVar = map[string]CommonChaincodeAction{
	invoke.Ping: Ping, // Проверка установки чейнкода и возможности записи в леджер
	// invoke.ClearLedger: ClearLedger, // Метод очищает все данные в ledger реализован в каждом чейнкоде
	// Configure - испосльзуется для инициализации информации в канале
}

var ChaincodeParameters ChaincodeParameter

var ChaincodeVar = CommonChaincode{
	Actions: CommonChaincodeActionVar,
}

func timeFuncFactory(adapter shim.ChaincodeStubInterface) func() (int64, error) {
	return func() (int64, error) {
		tt, err := adapter.GetTxTimestamp()
		if err != nil {
			return 0, err
		}
		return tt.Seconds, nil
	}
}

// Init initialize chaincode with proper parameters
func (this CommonChaincode) Init(adapter shim.ChaincodeStubInterface) (response pb.Response) {
	return shim.Success([]byte("Chaincode initialized successfully"))
}

func (this CommonChaincode) Configure(configArgs []string, adapter shim.ChaincodeStubInterface) error {
	lenArguments := len(configArgs)
	if lenArguments == 1 {
		var signedMessage requests.FullSignedRequest
		if err := json.Unmarshal([]byte(configArgs[0]), &signedMessage); err != nil {
			return GlobalLogger.Errorf(err.Error())
		}

		var systemParam *requests.Parameter
		if err := json.Unmarshal(signedMessage.Parameters.SystemParams, &systemParam); err != nil {
			return logging.Get().Errorf("CheckAccess: Unmarshaling error: %s", err.Error())
		}

		GlobalLogger.Infof("configArgs: %v", configArgs)

		err := json.Unmarshal(systemParam.Param, &ChaincodeParameters)
		if err != nil {
			return GlobalLogger.Errorf(err.Error())
		}

		err = adapter.PutState(invoke.Configure, systemParam.Param)
		if err != nil {
			return GlobalLogger.Errorf(err.Error())
		}

		GlobalLogger.Info("ChaincodeParameters	")
		GlobalLogger.Info(ChaincodeParameters.ToString())

	} else {
		return GlobalLogger.Errorf("Incorrect number of arguments. Expected 1 argument - ChaincodeParameter. But got %d", lenArguments)
	}

	return nil
}

// Invoke dispatch to the specific chaincode function based on the input parameters
func (this CommonChaincode) Invoke(adapter shim.ChaincodeStubInterface) (response pb.Response) {
	defer func() {
		GlobalLogger.Info("end invoke")
		if r := recover(); r != nil {
			errMsg := fmt.Sprintf("FATAL_ERROR: %f", r)
			response = shim.Error(errMsg)
			GlobalLogger.Error(errMsg)
		}
	}()
	functionName, args := adapter.GetFunctionAndParameters()
	GlobalLogger.Infof("ChannelName: %s", ChaincodeParameters.ChannelName)
	GlobalLogger.Infof("ChaincodeName: %s", ChaincodeParameters.ChaincodeName)
	GlobalLogger.Infof("functionName: %s", functionName)

	if functionName == invoke.Configure {
		defer func() {
			GlobalLogger.Info("end Configure")
			if r := recover(); r != nil {
				response = shim.Error(fmt.Sprintf("FATAL_ERROR: %f", r))
			}
		}()
		GlobalLogger.Info("Start configuring ..")

		if err := this.Configure(args, adapter); err != nil {
			GlobalLogger.Error(err.Error())
			return shim.Error(err.Error())
		}

		resultMsg := "Chaincode configured successfully"
		GlobalLogger.Infof(resultMsg)
		return shim.Success([]byte(resultMsg))
	} else {
		chaincodeParametersByteArray, err := adapter.GetState(invoke.Configure)
		if err != nil {
			GlobalLogger.Errorf(err.Error())
			return shim.Error(err.Error())
		}

		err = json.Unmarshal(chaincodeParametersByteArray, &ChaincodeParameters)
		if err != nil {
			GlobalLogger.Errorf(err.Error())
			return shim.Error(err.Error())
		}
	}

	timeGetter := timeFuncFactory(adapter)
	_, err := timeGetter()
	if err != nil {
		GlobalLogger.Error(err.Error())
		return shim.Error(fmt.Sprintf("Failed to get timestamp, failed due to %s", err))
	}

	GlobalLogger.Info("ChaincodeParameters")
	GlobalLogger.Info(ChaincodeParameters.ToString())

	result, err := this.InvokeChaincode(functionName, args, adapter, timeGetter)
	if err != nil {
		GlobalLogger.Error(err.Error())
		return shim.Error(err.Error())
	}

	GlobalLogger.Infof("Response: %s", result)
	bytes, err := json.Marshal(result)
	if err != nil {
		GlobalLogger.Error(err.Error())
		return shim.Error(err.Error())
	}

	return shim.Success(bytes)
}

// Invoke dispatch to the specific chaincode function based on the input parameters
func (this CommonChaincode) InvokeChaincode(functionName string, args []string, adapter shim.ChaincodeStubInterface, timeGetter TimeGetter) (interface{}, error) {
	if len(ChaincodeParameters.ChaincodeName) == 0 || len(ChaincodeParameters.ChannelName) == 0 {
		return nil, GlobalLogger.Error("ChaincodeParameters.ChannelName or ChaincodeParameters.ChaincodeName can't be nil")
	}
	iManager := GetManager(adapter, timeGetter)
	userParameter, systemParameter, err := CheckAccessParameterByArgsAndFunctionName(ChaincodeParameters.ChannelName, ChaincodeParameters.ChaincodeName, args, functionName, iManager)
	if err != nil {
		GlobalLogger.Error(err.Error())
		return nil, err
	}

	if userParameter != nil {
		GlobalLogger.Infof("userParameter.UserID: %s", userParameter.UserID)
		GlobalLogger.Infof("userParameter.Param: %s", userParameter.Param)
	}

	if systemParameter != nil {
		GlobalLogger.Infof("systemParameter.UserID: %s", systemParameter.UserID)
		GlobalLogger.Infof("systemParameter.Param: %s", systemParameter.Param)
	}

	// taking care to dispatch to the correct function based on the input parameter
	if action, ok := this.Actions[functionName]; ok {
		return action(userParameter, systemParameter, iManager)
	}

	// No action define with provided function name, reporting error
	err = GlobalLogger.Errorf("Invalid invoke function name, function name provided %s ", functionName)
	return nil, err
}
