package nums

import (
	"errors"
	"github.com/shopspring/decimal"
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
	"math"
	"strconv"
)

func NewAsset(symbol string, precision uint, maxValue string, minValue string) (*Asset, error) {
	mxv, err := decimal.NewFromString(maxValue)
	if err != nil {
		return nil, err
	}

	mnv, err := decimal.NewFromString(minValue)
	if err != nil {
		return nil, err
	}

	base := precision

	return &Asset{symbol, base, mxv, mnv}, nil
}

func (obj *Asset) String() string {
	return obj.CurrencyType
}

func (obj *Asset) Equals(o2 *Asset) bool {
	return obj.CurrencyType == o2.CurrencyType && obj.Scale == o2.Scale
}

func (obj *CurrencyNumber) Plain() PlainCurrencyNumber {
	return PlainCurrencyNumber{
		Amount:       obj.RoundBankString(),
		CurrencyType: obj.CurrencyType(),
	}
}

func (obj *PlainCurrencyNumber) IsAmountEmpty() (bool, error) {
	return IsAmountEmpty(obj.Amount)
}

func (obj *PlainCurrencyNumber) ValidateAmount() (error) {
	amountString := obj.Amount
	if len(amountString) == 0 {
		return errors.New("amount can't be empty string")
	}

	// TODO перенести проверку в этот метод
	if len(obj.CurrencyType) == 0 {
		return errors.New("Currency type can't by null or empty")
	}

	amount, err := decimal.NewFromString(amountString)
	if err != nil {
		return err
	}

	if amount.LessThan(decimal.Zero) {
		return logging.Get().Errorf("amount can't be negative number. actual: %s", amount.String())
	}

	return nil
}

func IsAmountEmpty(amountString string) (bool, error) {
	amount, err := decimal.NewFromString(amountString)
	if err != nil {
		return true, err
	}
	if amount.Equal(decimal.Zero) {
		return true, nil
	}

	return false, nil
}

func (obj *PlainCurrencyNumber) Validate() error {

	_, err := CurrencyTypeFrom(obj.CurrencyType)
	if err != nil {
		return err
	}

	err = obj.ValidateAmount()
	if err != nil {
		return err
	}

	return nil
}

func (obj *PlainCurrencyNumber) Number() (*CurrencyNumber, error) {
	err := obj.Validate()
	if err != nil {
		return nil, err
	}
	asset, _ := CurrencyTypeFrom(obj.CurrencyType)
	amount, _ := decimal.NewFromString(obj.Amount)

	currencyNumber := &CurrencyNumber{
		Asset: asset,
		Num:   amount,
	}
	return currencyNumber, nil
}

func (obj *PlainCurrencyNumber) RoundBankString() (*string, error) {
	currencyNumber, err := NumFromString(obj.Amount, obj.CurrencyType)
	if err != nil {
		return nil, err
	}
	roundBank := currencyNumber.RoundBankString()
	return &roundBank, nil
}

func (obj *CurrencyNumber) CurrencyType() string {
	return obj.Asset.String()
}

func (obj *CurrencyNumber) Copy() (*CurrencyNumber, error) {
	return NumFromString(obj.RoundBankString(), obj.Asset.CurrencyType)
}

// сравнение типов валют у двух любых CurrencyNumber
// если обе валюты nil, тогда типы одинаковые
func (obj *CurrencyNumber) AssetEquals(b *CurrencyNumber) error {
	if obj.Asset == nil || b.Asset == nil {
		return logging.Get().Errorf("can't equals nil assets")
	}

	if obj.Asset.Equals(b.Asset) {
		return nil
	} else {
		return logging.Get().Errorf("Assets are different: %s vs %s", obj.CurrencyType(), b.CurrencyType())
	}
}

func (obj *CurrencyNumber) UpdateAsset(asset *Asset) *CurrencyNumber {
	obj.Asset = asset
	return obj
}

func (obj *CurrencyNumber) RoundBankString() string {
	Num := obj.Num
	i := int32(obj.Asset.Scale)
	roundBank := Num.RoundBank(i)
	fixed := roundBank.StringFixed(i)
	return fixed
}

func FinRound(n float64, scale uint) float64 {
	offset := math.Pow(10, float64(scale))
	n *= offset

	x, d := math.Modf(n)

	if d == 0.5 {
		if (int(x)+1)%2 == 0 {
			n = math.Ceil(n)
		} else {
			n = math.Floor(n)
		}
	} else if d > 0.5 {
		n = math.Ceil(n)
	} else {
		n = math.Floor(n)
	}

	return n / offset
}

func MinNumber(a *CurrencyNumber, b *CurrencyNumber) *CurrencyNumber {
	if a.Num.Cmp(b.Num) < 0 {
		return a
	} else {
		return b
	}
}

func MinUInt(a uint64, b uint64) uint64 {
	return uint64(math.Min(float64(a), float64(b)))
}

// валидация валюты кол-во символов после запятой, максимальное и минимальное значение
func (res *CurrencyNumber) Validate() (*CurrencyNumber, error) {
	plainCurrencyNumber := res.Plain()
	err := plainCurrencyNumber.Validate()
	if err != nil {
		return nil, err
	}

	flag, err := decimal.NewFromString("-1")
	if err != nil {
		return nil, err
	}

	if res.Num.LessThan(decimal.Zero) ||
		((res.Num.GreaterThan(res.Asset.MaxValue) && !res.Asset.MaxValue.Equals(flag)) ||
		(res.Num.LessThan(res.Asset.MinValue) && !res.Asset.MinValue.Equals(flag))) &&
		!res.Num.Equals(decimal.Zero) {

		//assetMinValue := res.Asset.MinValue.RoundBank(int32(res.Asset.Scale)).StringFixed(int32(res.Asset.Scale))
		//assetMaxValue := res.Asset.MaxValue.RoundBank(int32(res.Asset.Scale)).StringFixed(int32(res.Asset.Scale))
		//assetScale := res.Num.StringFixed(int32(res.Asset.Scale))
		err = logging.Get().Errorf("Value is out of bounds [min: %s, max: %s, scale: %d] but got %s",
			res.Asset.MinValue, res.Asset.MaxValue, res.Asset.Scale, res.Num.String())
		if err != nil {
			return nil, err
		}
		res.Num = decimal.Zero
	}

	return res, nil
}

// во время преобразования amount и currencyType
// происходит проверка типа валюты
// происходит проверка то что число amount не пусто и не меньше 0 включая 0
// также amount проверяется по таблице ограничений для валюты
func NumFromString(amount string, currencyType string) (*CurrencyNumber, error) {
	plainCurrencyNumber := PlainCurrencyNumber{
		Amount:amount,
		CurrencyType:currencyType,
	}
	number, err := plainCurrencyNumber.Number()
	if err != nil {
		return nil, err
	}
	return number.Validate()
}

// TODO вынести в тестовый пакет

// FIXME use only for test
func NumFromFloat(n float64, currencyType string) (*CurrencyNumber, error) {
	string := strconv.FormatFloat(n, 'f', 6, 64)
	return NumFromString(string, currencyType)
}

// FIXME use only for test
func NumFromInt(n uint64, currencyType string) (*CurrencyNumber, error) {
	return NumFromString(strconv.FormatUint(n, 10), currencyType)
}

// FIXME use only for test
func FormatFloat(f float64, scale uint) string {
	return strconv.FormatFloat(f, 'f', int(scale), 64)
}

func One(n *CurrencyNumber) (*CurrencyNumber, error) {
	return NumFromFloat(1, n.Asset.CurrencyType)
}